@@ -1,4 +1,6 @@
 #!/usr/bin/python
+# SPDX-License-Identifier: AGPL-3.0-or-later
+from __future__ import print_function
 from .version import __version__
 
 import getpass
@@ -59,6 +61,7 @@ def parser_opt_args():
     group.add_option('--mr-accept', help='Auto-accept meeting requests', **_bool())
     group.add_option('--mr-accept-conflicts', help='Auto-accept conflicting meeting requests', **_bool())
     group.add_option('--mr-accept-recurring', help='Auto-accept recurring meeting requests', **_bool())
+    group.add_option('--mr-process', help='Auto-process meeting requests', **_bool())
     parser.add_option_group(group)
 
     group = OptionGroup(parser, "Send-as and delegations", "")
@@ -118,7 +121,7 @@ UPDATE_MATRIX = {
     ('name',): ('companies', 'groups', 'users'),
     ('email', 'add_sendas', 'remove_sendas'): ('users', 'groups'),
     ('fullname', 'password', 'password_prompt', 'admin_level', 'active', 'reset_folder_count'): ('users',),
-    ('mr_accept', 'mr_accept_conflicts', 'mr_accept_recurring', 'hook_archive', 'unhook_archive'): ('users',),
+    ('mr_accept', 'mr_accept_conflicts', 'mr_accept_recurring', 'hook_archive', 'unhook_archive', 'mr_process'): ('users',),
     ('ooo', 'ooo_clear', 'ooo_subject', 'ooo_message', 'ooo_from', 'ooo_until'): ('users',),
     ('add_feature', 'remove_feature', 'add_delegation', 'remove_delegation'): ('users',),
     ('send_only_to_delegates',): ('users',),
@@ -134,7 +137,7 @@ UPDATE_MATRIX = {
 if sys.hexversion >= 0x03000000:
     def _encode(s):
         return s
-else:
+else: # pragma: no cover
     def _encode(s):
         return s.encode(sys.stdout.encoding or 'utf8')
 
@@ -157,9 +160,9 @@ def name_flags(s):
 def list_users(intro, users):
     users = list(users)
     print(intro + ' (%d):' % len(users))
-    fmt = '{:>16}{:>20}{:>20}{:>40}'
+    fmt = '{0:>16} {1:>20} {2:>20} {3:>40}'
     print(fmt.format('User', 'Full Name', 'Homeserver', 'Store'))
-    print(96*'-')
+    print(99*'-')
     for user in sorted(users, key=lambda u: u.name):
         print(fmt.format(_encode(user.name), _encode(user.fullname), _encode(user.home_server), user.store.guid if user.store else ''))
     print
@@ -167,7 +170,7 @@ def list_users(intro, users):
 def list_groups(intro, groups):
     groups = list(groups)
     print(intro + ' (%d):' % len(groups))
-    fmt = '\t{:>16}'
+    fmt = '\t{0:>16}'
     print(fmt.format('Groupname'))
     print('\t'+16*'-')
     for group in sorted(groups, key=lambda g: g.name):
@@ -177,17 +180,17 @@ def list_groups(intro, groups):
 def list_companies(intro, companies):
     companies = list(companies)
     print(intro + ' (%d):' % len(companies))
-    fmt = '\t{:>32}{:>32}'
+    fmt = '\t{0:>32} {1:>32}'
     print(fmt.format('Companyname', 'System Administrator'))
-    print('\t'+64*'-')
+    print('\t'+65*'-')
     for company in sorted(companies, key=lambda c: c.name):
         print(fmt.format(_encode(company.name), _encode(company.admin.name) if company.admin else ''))
 
 def list_orphans(server):
     print('Stores without users:')
-    fmt = '\t{:>32}{:>20}{:>16}{:>16}{:>16}'
+    fmt = '\t{0:>32} {1:>20} {2:>16} {3:>16} {4:>16}'
     print(fmt.format('Store guid', 'Username', 'Last login', 'Store size', 'Store type'))
-    print('\t'+100*'-')
+    print('\t'+104*'-')
     for store in server.stores():
         if store.orphan:
             username = store.user.name if store.user else ''
@@ -212,9 +215,9 @@ def user_counts(server): # XXX allowed/available
     if sys.hexversion >= 0x03000000: # XXX shouldn't be necessary
         stats = dict([(s.decode('ascii'), stats[s].decode('ascii')) for s in stats])
     print('User counts:')
-    fmt = '\t{:>12}{:>10}'
+    fmt = '\t{0:>12} {1:>10}'
     print(fmt.format('', 'Used'))
-    print('\t'+42*'-')
+    print('\t'+43*'-')
     print(fmt.format('Active', stats['usercnt_active']))
     print(fmt.format('Non-active', stats['usercnt_nonactive']))
     print(fmt.format('NA Users', stats['usercnt_na_user']))
@@ -222,8 +225,14 @@ def user_counts(server): # XXX allowed/available
     print(fmt.format('NA Equipment', stats['usercnt_equipment']))
     print(fmt.format('Total', int(stats['usercnt_active'])+int(stats['usercnt_nonactive'])))
 
+def quota_str(limit): # TODO in pyko?
+    if not limit:
+        return 'unlimited'
+    else:
+        return '%d MB' % (float(limit) / 2**20)
+
 def user_details(user):
-    fmt = '{:<30}{:<}'
+    fmt = '{0:<30} {1:<}'
     print(fmt.format('Name:', _encode(user.name)))
     print(fmt.format('Full name:', _encode(user.fullname)))
     print(fmt.format('Email address:', user.email))
@@ -250,6 +259,8 @@ def user_details(user):
             print(fmt.format('    Accept conflicting:', yesno(user.autoaccept.conflicts)))
             print(fmt.format('    Accept recurring:', yesno(user.autoaccept.recurring)))
 
+        print(fmt.format('Auto-process meeting request:', yesno(user.autoprocess.enabled)))
+
         ooo = 'disabled'
         if user.outofoffice.enabled:
             ooo = user.outofoffice.period_desc + (' (currently %s)' % ('active' if user.outofoffice.active else 'inactive'))
@@ -257,15 +268,15 @@ def user_details(user):
 
     print('Current user store quota settings:')
     print(fmt.format('    Quota overrides:', yesno(not user.quota.use_default)))
-    print(fmt.format('    Warning level:', str(user.quota.warning_limit or 'unlimited')))
-    print(fmt.format('    Soft level:', str(user.quota.soft_limit or 'unlimited')))
-    print(fmt.format('    Hard level:', str(user.quota.hard_limit or 'unlimited')))
+    print(fmt.format('    Warning level:', quota_str(user.quota.warning_limit)))
+    print(fmt.format('    Soft level:', quota_str(user.quota.soft_limit)))
+    print(fmt.format('    Hard level:', quota_str(user.quota.hard_limit)))
 
     list_groups('Groups', user.groups())
     list_permissions(user.store)
 
 def group_details(group):
-    fmt = '{:<16}{:<}'
+    fmt = '{0:<16} {1:<}'
     print(fmt.format('Name:', _encode(group.name)))
     print(fmt.format('Email address:', group.email))
     print(fmt.format('Address Book:', ('hidden' if group.hidden else 'visible')))
@@ -273,7 +284,7 @@ def group_details(group):
     list_users('Users', group.users())
 
 def company_details(company, server):
-    fmt = '{:<30}{:<}'
+    fmt = '{0:<30} {1:<}'
     print(fmt.format('Name:', _encode(company.name)))
     if company.admin:
         print(fmt.format('Sysadmin:', _encode(company.admin.name)))
@@ -290,7 +301,7 @@ def company_details(company, server):
     list_permissions(company.public_store)
 
 def store_details(store, server):
-    fmt = '{:<30}{:<}'
+    fmt = '{0:<30} {1:<}'
     print(fmt.format('GUID:', store.guid))
     print(fmt.format('Size:', '%.2f MB' % (store.size / 2**20)))
     list_permissions(store)
@@ -310,11 +321,11 @@ def quota_options(user, options):
     if options.quota_override is not None:
         user.quota.use_default = not options.quota_override
     if options.quota_warn is not None:
-        user.quota.warning_limit = options.quota_warn
+        user.quota.warning_limit = options.quota_warn * (2**20) # TODO make pyko accept '2 GB' etc?
     if options.quota_soft is not None:
-        user.quota.soft_limit = options.quota_soft
+        user.quota.soft_limit = options.quota_soft * (2**20)
     if options.quota_hard is not None:
-        user.quota.hard_limit = options.quota_hard
+        user.quota.hard_limit = options.quota_hard * (2**20)
 
 def delegation_options(user, options, server):
     for delegation in options.add_delegation:
@@ -363,8 +374,6 @@ def user_options(name, options, server):
 
     shared_options(user, options, server)
 
-    if options.create_store:
-        user.create_store()
     if options.unhook_store:
         user.unhook()
     if options.hook_store:
@@ -373,6 +382,8 @@ def user_options(name, options, server):
         user.unhook_archive()
     if options.hook_archive:
         user.hook_archive(server.store(options.hook_archive))
+    if options.create_store:
+        user.create_store()
 
     if options.reset_folder_count:
         for folder in [user.root] + list(user.folders()):
@@ -385,6 +396,9 @@ def user_options(name, options, server):
     if options.mr_accept_recurring is not None:
         user.autoaccept.recurring = options.mr_accept_recurring
 
+    if options.mr_process is not None:
+        user.autoprocess.enabled = options.mr_process
+
     for feature in options.add_feature:
         user.add_feature(feature)
     for feature in options.remove_feature:
@@ -446,12 +460,12 @@ def group_options(name, options, server):
         server.delete(group)
 
 def company_update_options(company, options, server):
-    if options.create_store:
-        company.create_public_store()
     if options.unhook_store:
         company.unhook_public_store()
     if options.hook_store:
         company.hook_public_store(server.store(options.hook_store))
+    if options.create_store:
+        company.create_public_store()
 
     shared_options(company, options, server)
 
@@ -550,14 +564,10 @@ def check_options(parser, options, server):
     updates = [name for name in sum(UPDATE_MATRIX, ()) if getattr(options, name) not in (None, [])]
 
     if not (actions or updates or objtypes):
-        return False
+        raise Exception('no object or action specified')
     if not (actions or updates):
         options.details = True
 
-    if not (objtypes or actions or updates):
-        parser.print_help()
-        sys.exit(1)
-
     for opts, legaltypes in list(ACTION_MATRIX.items()) + list(UPDATE_MATRIX.items()):
         for opt in opts:
             if getattr(options, opt) not in (None, []):
@@ -572,8 +582,6 @@ def check_options(parser, options, server):
             if getattr(options, opt) is not None:
                 raise Exception('%s option requires --company for multitenant setup' % orig_option(opt))
 
-    return True
-
 def main():
     try:
         parser, options, args = parser_opt_args()
@@ -595,10 +603,10 @@ def main():
 
     except Exception as e:
         if 'options' in locals() and options.debug:
-            print(traceback.format_exc())
+            print(traceback.format_exc(), file=sys.stderr)
         else:
-            print(_encode(str(e)))
+            print(_encode(str(e)), file=sys.stderr)
         sys.exit(1)
 
 if __name__ == '__main__':
-    main()
+    main() # pragma: no cover
