@@ -1,4 +1,5 @@
 #!@PYTHON@
+# SPDX-License-Identifier: AGPL-3.0-or-later
 from datetime import datetime, timedelta
 import sys
 import kopano
@@ -6,7 +7,7 @@ import kopano
 if sys.hexversion >= 0x03000000:
     def _decode(s):
         return s
-else:
+else: # pragma: no cover
     def _decode(s):
         return s.decode(getattr(sys.stdin, 'encoding', 'utf8') or 'utf8')
 
@@ -29,4 +30,4 @@ def main():
     user.freebusy.publish(now - timedelta(7), now + timedelta(180))
 
 if __name__ == '__main__':
-    main()
+    main() # pragma: no cover
